(ns puppetlabs.rbac.db-utils
  (:require [clojure.java.jdbc :as jdbc]
            [migratus.core :as migratus]))

(def migratus-cfg {:store :database
                   :migration-dir "puppetlabs/rbac/migrations"})

(defn public-tables
  "Get the names of all public tables in a database"
  [db-spec]
  (let [query "SELECT table_name FROM information_schema.tables WHERE LOWER(table_schema) = 'public'"
               results (jdbc/query db-spec [query])]
    (map :table_name results)))

(defn drop-public-tables!
  "Drops all public tables in a database. Super dangerous."
  [db-spec]
  (when-let [tables (seq (public-tables db-spec))]
    (apply jdbc/db-do-commands db-spec (map #(format "DROP TABLE %s CASCADE" %) (seq tables)))))

(defn mcfg
  "Small helper that takes a db-spec and merges it with the (static
   and default) migratus config"
  [db-spec]
  (merge migratus-cfg {:db db-spec}))

(defn- get-last-migration
  "Returns the most recent migration ID"
  [db-spec]
  (-> (jdbc/query db-spec ["SELECT id FROM schema_migrations ORDER BY id DESC LIMIT 1"])
      first
      :id))

(defn migrate!
  "Migrate a database. Takes the path of a config file."
  [db-spec]
  (migratus/migrate (mcfg db-spec)))

(defn migrate-down!
  "Migrate backwards. Migrates one migration at a time unless optional
   id is provided."
  [db-spec & [migration-id]]
  (let [migration-id (if (number? migration-id) migration-id (get-last-migration db-spec))]
    (migratus/down (mcfg db-spec) migration-id)))

(defn migrate-reset!
  "Downs every migration."
  [db-spec]
  (let [ids (map :id (jdbc/query db-spec ["SELECT id FROM schema_migrations ORDER BY id DESC"]))]
    (doseq [id ids]
      (migrate-down! db-spec id))))

(defn migrate-reload!
  "Downs all migrations and then ups them all."
  [db-spec]
  (migrate-reset! db-spec)
  (migrate! db-spec))
